/**
 * Types an optional type at TypeScript check level.
 */
type Option<T> = T | undefined;
/**
 * Creates an [Option] that contains a value.
 *
 * @param value - the contained value
 */
declare const Some: <T>(value: T) => T;
/**
 * Creates an empty [Option].
 */
declare const None: <T>() => undefined;
/**
 * Type-guard predicate that checks if an [Option] contains a value.
 *
 * @param value - the [Option] value to check.
 */
declare const present: <T>(value: Option<T>) => value is T;
/**
 * Type-guard predicate that checks if an [Option] does not contain a value.
 *
 * @param value - the [Option] value to check.
 */
declare const empty: <T>(value: Option<T>) => value is undefined;

/**
 * Types an Either monad. As per convention, the [L] type represents a failure, while [R] the success type.
 */
type Either<L, R> = Left<L> | Right<R>;
/**
 * Types the left side of an [Either] monad, typically used to describe failures.
 */
type Left<L> = {
    readonly tag: 'left';
    readonly left: L;
};
/**
 * Creates the left hand of an [Either].
 *
 * @param value - the value that represents the left hand
 */
declare const Left: <L, R>(value: L) => {
    tag: "left";
    left: L;
};
/**
 * Types the right side of an [Either] monad, typically used to describe success results.
 */
type Right<R> = {
    readonly tag: 'right';
    readonly right: R;
};
/**
 * Creates the right hand of an [Either].
 *
 * @param value - the value that represents the right hand
 */
declare const Right: <L, R>(value: R) => {
    tag: "right";
    right: R;
};
/**
 * Type-guard predicate that checks if an [Either] is left-handed.
 *
 * @param value - the [Either] monad to check.
 */
declare const isLeft: <L, R>(value: Either<L, R>) => value is Left<L>;
/**
 * Type-guard predicate that checks if an [Either] is right-handed.
 *
 * @param value - the [Either] monad to check.
 */
declare const isRight: <L, R>(value: Either<L, R>) => value is Right<R>;
/**
 * Folds an [Either] monad by computing the left callback, if the monad is left-handed or the right callback if a right-handed.
 *
 * @param value - the monad value to fold
 * @param left - callback for folding the left-hand
 * @param right - callback for folding the right-hnd
 * @returns the folded value based on the computed callbacks.
 */
declare const fold: <L, R, F>(value: Either<L, R>, left: (l: L) => F, right: (r: R) => F) => F;

/**
 * Types a callback that computes a value.
 */
type LazyValueCallback<T> = () => T;
/**
 * Types a value that either needs to be computed or has been computed already.
 */
type Lazy<T> = T | LazyValueCallback<T>;
/**
 * Creates a [Lazy] based on a synchronous callback.
 *
 * @param callback - the callback that computes a value.
 */
declare const Lazy: <T>(callback: LazyValueCallback<T>) => Lazy<T>;
/**
 * Creates a [Lazy] based on an asynchronous callback.
 *
 * @param callback - the callback that computes a value.
 */
declare const AsyncLazy: <T>(callback: LazyValueCallback<Promise<T>>) => LazyValueCallback<Promise<T>>;
/**
 * Type-guard predicate that checks if a [Lazy] has been computed already.
 *
 * @param value - the [Lazy] value to check.
 */
declare const computed: <T>(value: Lazy<T>) => value is T;
/**
 * Computes a [Lazy] value if it hasn't been already.
 *
 * @param value - the computed value.
 */
declare const compute: <T>(value: Lazy<T>) => T;

/**
 * Types an error with all three core properties visible: type (name), cause (message) and stack (trace).
 */
type TypedError = {
    type: string;
    cause: string;
    stack: string | undefined;
};
/**
 * Creates a new [TypedError] with the option of inferring the type value from a specific [TypedError].
 *
 * @param type - the identifier of the error
 * @param cause - the message that describes the error
 * @param stack - the trace of the execution stack up until the error origin.
 * @returns an object of type [TypedError]
 */
declare const TypedError: <T extends TypedError>(type: T["type"], cause: string, stack?: string) => {
    type: T["type"];
    cause: string;
    stack: string | undefined;
};
/**
 * Type-guard predicate that checks if a [TypedError] is of a specific type.
 *
 * @param value - the [TypedError] value to check
 * @param type - the type literal that asserts if the value is of required type.
 */
declare const isTypedError: <T extends TypedError>(value: TypedError, type: T["type"]) => value is T;
/**
 * Wraps a [TypedError] in a [Error].
 *
 * @param value - the [TypedError] to wrap
 * @returns the wrapped error
 */
declare const wrap: (value: TypedError) => Error;
/**
 * Unwraps an [Error] in a [TypedError].
 *
 * @param value - the [Error] to unwrap
 * @returns the unwrapped error
 */
declare const unwrap: (value: Error) => {
    cause: string;
    type: string;
    stack: string | undefined;
};

/**
 * Executes a callback in a safe throwable environment, by catching errors and
 * returning them in a [Either<L, R>] value.
 *
 * @param call - the callback to execute
 * @param onError - a callback to transform an internal [Error] in a [TypedError]
 */
declare const safeThrowCall: <L extends TypedError, R>(call: () => Either<L, R>, onError?: ((error: Error) => L) | undefined) => Either<L, R> | {
    tag: "left";
    left: {
        type: "unknown-error";
        cause: string;
        stack: string | undefined;
    } | L;
};
/**
 * Sane as [safeThrowCall] but for an asynchronous callback.
 *
 * @param call - the asynchronous callback to execute
 * @param onError - a callback to transform an internal [Error] in a [TypedError]
 */
declare const safeAsyncThrowCall: <L extends TypedError, R>(call: () => Promise<Either<L, R>>, onError?: ((error: Error) => L) | undefined) => Promise<Either<L, R> | {
    tag: "left";
    left: {
        type: "unknown-error";
        cause: string;
        stack: string | undefined;
    } | L;
}>;

/**
 * Types a [TypedError] for IO operations.
 */
type IOError = ReadError | WriteError;
/**
 * A [TypedError] for IO read operations (like reading bytes from a stream).
 */
type ReadError = {
    readonly type: 'read-error';
} & TypedError;
/**
 * Creates a [IOError] for a faulty read operation.
 *
 * @param cause - what caused the read error
 * @param stack - the stack trace until the error origin
 */
declare const ReadError: (cause: string, stack?: string) => {
    type: "read-error";
    cause: string;
    stack: string | undefined;
};
/**
 * A [TypedError] for IO write operations (like writing bytes from a stream).
 */
type WriteError = {
    readonly type: 'write-error';
} & TypedError;
/**
 * Creates a [IOError] for a faulty write operation.
 *
 * @param cause - what caused the read error
 * @param stack - the stack trace until the error origin
 */
declare const WriteError: (cause: string, stack?: string) => {
    type: "write-error";
    cause: string;
    stack: string | undefined;
};
/**
 * Type-guard predicate that checks if an [IOError] is of type [ReadError].
 *
 * @param value - the error value to check
 */
declare const isReadError: (value: IOError) => boolean;
/**
 * Type-guard predicate that checks if an [IOError] is of type [WriteError].
 *
 * @param value - the error value to check
 */
declare const isWriteError: (value: IOError) => boolean;

/**
 * A [TypedError] for unknown errors.
 */
type UnknownError = {
    readonly type: 'unknown-error';
} & TypedError;
/**
 * Standard call to create an [UnknownError].
 *
 * @param cause - what caused the error
 * @param stack - the stack trace until the error origin
 */
declare const UnknownError: (cause: string | unknown, stack?: string) => {
    type: "unknown-error";
    cause: string;
    stack: string | undefined;
};
/**
 * Converts a value of type [unknown] to [Error].
 *
 * @param error - the unknown value
 */
declare const safeUnknownError: (error: unknown) => Error;

export { AsyncLazy, Either, IOError, Lazy, Left, None, Option, ReadError, Right, Some, TypedError, UnknownError, WriteError, compute, computed, empty, fold, isLeft, isReadError, isRight, isTypedError, isWriteError, present, safeAsyncThrowCall, safeThrowCall, safeUnknownError, unwrap, wrap };
